// img2avi 0.01
// Copyright 2011. Claybird <claybird.without.wing@gmail.com>
//                 http://claybird.sakura.ne.jp/
// License: New BSD License
// 
// make an uncompressed avi from various formats of images
// code requirements:
//    GDI+:
//        reading images
//    Video for Windows:
//        making avi
//    Active Template Library(ATL) & Windows Template Library(WTL):
//        wrapper library for GDI+
//        file utilities
//
//

#define _WTL_NO_CSTRING	//disable WTL::CString
#define _WTL_NO_WTYPES	//disable WTL::CPoint,WTL::CSize,WTL::CRect


#include <stdio.h>
#include <windows.h>
#include <vfw.h>
#include <list>
#include <vector>
#include <gdiplus.h>	//Image loader

#include <atlstr.h>
#include <atlbase.h>
#include <atlapp.h>
#include <atlmisc.h>
#include <atlimage.h>	//CImage
#include <atlpath.h>

struct OPTIONS
{
	CString strOutputFile;
	bool bVerbose;
	int nFramePerSec;
};

// find files in the current directory
// lpFormat="*.bmp" or "exam_*.bmp"
// lpFormat=NULL then all files
bool UtilEnumFiles(std::list<CString> &rFileList,LPCTSTR lpFormat)
{
	CFindFile cFindFile;

	CString strParam="*";
	if(lpFormat)strParam=lpFormat;

	BOOL bContinue=cFindFile.FindFile(strParam);
	while(bContinue){
		if(!cFindFile.IsDots()){
			if(!cFindFile.IsDirectory()){
				rFileList.push_back(cFindFile.GetFilePath());
			}
		}
		bContinue=cFindFile.FindNextFile();
	}

	return !rFileList.empty();
}

void ShowUsage()
{
	printf(
		"img2avi Ver.0.01\n"
		"Make AVI file from image files using GDI+ and Video For Windows\n"
		"Copyright 2011. Claybird <claybird.without.wing@gmail.com>\n"
		"Usage:\n"
		"img2avi [Options] <OutputFile> <ImageFile> [<ImageFile>...]\n"
		"Wildcards ('*' or '?') are accepted for <ImageFile>\n"
		"[Options]\n"
		"-f=<n>\t:frames per sec\n"
		"-v\t:verbose output\n"
		"-s\t:sort by filename\n"
		"-h\t:shows this help\n"
		);
}

// Command line parser
bool ParseOption(OPTIONS& option,std::list<CString>& fileList,int argc,char** argv)
{
	option.strOutputFile="";
	option.bVerbose=false;
	bool bSort=false;	// true if sort by filename is required
	option.nFramePerSec=25;

	for(int i=1;i<argc;i++){
		if('-'==argv[i][0]){
			switch(argv[i][1]){
			case 'h':
				ShowUsage();
				return false;
			case 'v':	//Verbose output
				option.bVerbose=true;
				break;
			case 's':	//Sort
				bSort=true;
				break;
			case 'f':	//Frame per sec
				if(argv[i][2]=='='){
					option.nFramePerSec=atoi(argv[i]+3);
					if(option.nFramePerSec<=0){
						printf("Error:zero or negative frame per sec is not accepted\n");
						return false;
					}
				}else{
					printf("Error:frame per sec is required for -f option\n");
					return false;
				}
				break;
			default:
				printf("Warning:Unknown option %s\n",argv[i]);
			}
		}else if(option.strOutputFile.IsEmpty()){	//output filename
			option.strOutputFile=argv[i];
		}else{	//image files
			CString strEnum=argv[i];
			if(-1!=strEnum.Find("*") || -1!=strEnum.Find("?")){
				//if wildcards found, look for files
				std::list<CString> tmpList;
				if(!UtilEnumFiles(tmpList,strEnum)){
					printf("No file matched for %s\n",(LPCTSTR)strEnum);
				}
				//sort by filename
				tmpList.sort();
				fileList.insert(fileList.end(),tmpList.begin(),tmpList.end());
			}else{
				//raw filename
				fileList.push_back(strEnum);
			}
		}
	}

	if(option.strOutputFile.IsEmpty() || fileList.empty()){
		ShowUsage();
		return false;
	}

	if(bSort){
		fileList.sort();
	}
	return true;
}

bool Load24BitDIB(CImage& rImage,const char* fname)
{
	CImage bmSrc;
	if(FAILED(bmSrc.Load(fname))){
		return false;
	}

	if(bmSrc.GetBPP()==24){
		//move image
		rImage.Attach(bmSrc.Detach());
		bmSrc.Destroy();
	}else{
		//convert to 24bit DIB
		rImage.Create(bmSrc.GetWidth(),bmSrc.GetHeight(),24);
		bmSrc.BitBlt(rImage.GetDC(),0,0);
		rImage.ReleaseDC();
	}
	return true;
}


int main(int argc,char** argv)
{
	OPTIONS option;
	std::list<CString> fileList;
	if(!ParseOption(option,fileList,argc,argv))return 2;

	//CImage class initializes GDI+
	//ULONG_PTR token;
	//Gdiplus::GdiplusStartupInput gdiplusStartupInput;
	//Gdiplus::GdiplusStartup(&token,&gdiplusStartupInput,NULL);

	//---Image properties
	BITMAPINFOHEADER bmih={sizeof(BITMAPINFOHEADER)};
	{
		CImage cImage;
		if(!Load24BitDIB(cImage,fileList.front())){
			printf("Failed to open file %s\n",(LPCTSTR)fileList.front());
			return 1;
		}

		bmih.biWidth=cImage.GetWidth();
		bmih.biHeight=cImage.GetHeight();
		bmih.biPlanes=1;
		bmih.biBitCount=cImage.GetBPP();
		bmih.biCompression=BI_RGB;
		bmih.biSizeImage=abs(cImage.GetPitch()*cImage.GetHeight());
		cImage.Destroy();


		if(option.bVerbose){	//Status report
			printf("Output format: %d x %d, %dbits\n",bmih.biWidth,bmih.biHeight,bmih.biBitCount);
		}
	}


	//---Preprocess for making AVI
	//Stream
	AVISTREAMINFO si={0};
	si.fccType=streamtypeVIDEO;
	si.fccHandler=comptypeDIB;
	si.dwScale=1;
	//Frames per sec
	si.dwRate=option.nFramePerSec;
	si.dwLength=(int)fileList.size();
	si.dwQuality=-1;
	//Image size
	si.rcFrame=CRect(0,0,bmih.biWidth,bmih.biHeight);

	//Delete existing avi file
	DeleteFile(option.strOutputFile);

	AVIFileInit();

	PAVIFILE pavi=NULL;
	PAVISTREAM pstm=NULL;

	if(AVIFileOpen(&pavi,option.strOutputFile,OF_CREATE | OF_WRITE | OF_SHARE_DENY_NONE,NULL)!=0){
		printf("Error:Failed to open AVI file %s\n",(LPCTSTR)option.strOutputFile);
		return 1;
	}
	if(AVIFileCreateStream(pavi,&pstm,&si)!=0){
		printf("Error:Failed to create stream\n");
		return 1;
	}
	if(AVIStreamSetFormat(pstm,0,&bmih,sizeof(bmih))!=0){
		printf("Error:Failed to set stream format\n");
		return 1;
	}

	//Write to AVI
	bool bOK=true;
	int idxFrame=0;
	for(std::list<CString>::const_iterator ite=fileList.begin();ite!=fileList.end();++ite,idxFrame++){
		CImage cImage;
		if(option.bVerbose){	//Verbose output
			printf("#%d:Processing %s\n",idxFrame,(LPCTSTR)(*ite));
		}
		if(!Load24BitDIB(cImage,*ite)){
			printf("Failed to open file %s\n",(LPCTSTR)(*ite));
			bOK=false;
			cImage.Destroy();
			break;
		}

		//Check image size
		if(cImage.GetWidth()!=bmih.biWidth || cImage.GetHeight()!=bmih.biHeight){
			printf("Error:image size of file %s does not match(%dx%d)\n",(LPCTSTR)(*ite),cImage.GetWidth(),cImage.GetHeight());
			bOK=false;
			cImage.Destroy();
			break;
		}

		void *pData=cImage.GetBits();
		if(cImage.GetPitch()<0){
			//NOTE:Upside down: pData poins at the head of the last line of the image
			//move the pointer to the head
			pData=((BYTE*)cImage.GetBits())-abs(cImage.GetPitch()*(cImage.GetHeight()-1));
		}

		int size=abs(cImage.GetPitch()*cImage.GetHeight());
		//write a frame
		if(AVIStreamWrite(pstm,idxFrame,1,pData,size,AVIIF_KEYFRAME,NULL,NULL)!=0){
			printf("Failed to write file %s to avi stream\n",(LPCTSTR)(*ite));
			bOK=false;
			break;
		}

		cImage.Destroy();
	}

	//End output
	AVIStreamRelease(pstm);
	AVIFileRelease(pavi);
	AVIFileExit();

	if(bOK){
		//status report
		printf("Total %d frames, %.1f sec\n",si.dwLength,si.dwLength/(float)si.dwRate);
	}else{
		DeleteFile(option.strOutputFile);
	}

	//clean up of GDI+
	//Gdiplus::GdiplusShutdown(token);
	return 0;
}
